c
c
      subroutine linv3f (a,b,ijob,n,ia,d1,d2,wkarea,ier)                
c
c   imsl routine name   - linv3f                                        
c                                                                       
c-----------------------------------------------------------------------
c                                                                       
c   computer            - prim77/single                                 
c                                                                       
c   latest revision     - june 1, 1982                                  
c                                                                       
c   purpose             - in place inverse, equation solution, and/or   
c                           determinant evaluation - full storage mode  
c                                                                       
c   usage               - call linv3f (a,b,ijob,n,ia,d1,d2,wkarea,ier)  
c                                                                       
c   arguments    a      - input/output matrix of dimension n by n. see  
c                           parameter ijob.                             
c                b      - input/output vector of length n when ijob =   
c                           2 or 3. otherwise, b is not used.           
c                         on input, b contains the right hand side of   
c                           of the equation ax = b.                     
c                         on output, the solution x replaces b.         
c                ijob   - input option parameter. ijob = i implies:     
c                           i = 1, invert matrix a. a is replaced by    
c                             its inverse.                              
c                           i = 2, solve the equation ax = b. a is      
c                             replaced by the lu decomposition of a     
c                             rowwise permutation of a, where u is      
c                             upper triangular and l is lower           
c                             triangular with unit diagonal.            
c                             the unit diagonal of l is not stored.     
c                           i = 3, solve ax = b and invert matrix a.    
c                             a is replaced by its inverse.             
c                           i = 4, compute the determinant of a.        
c                             a is replaced by the lu decomposition     
c                             of a rowwise permutation of a.            
c                n      - order of a. (input)                           
c                ia     - row dimension of matrix a exactly as          
c                           specified in the dimension statement in the 
c                           calling program. (input)                    
c                d1     - input/output. if the d1 and d2 components     
c                d2         of determinant(a) = d1*2**d2 are desired,   
c                           input d1.ge.0. otherwise, input d1.lt.0.    
c                           d2 is never input.                          
c                wkarea - work area of length at least 2*n for ijob=1   
c                           or ijob=3.                                  
c                         work area of length at least n for ijob=2     
c                           or ijob=4.                                  
c                ier    - error parameter. (output)                     
c                         warning with fix                              
c                           ier = 65 indicates that ijob was less than  
c                             1 or greater than 4. ijob is assumed to   
c                             be 4.                                     
c                         terminal error                                
c                           ier = 130 indicates that matrix a is        
c                             algorithmically singular. (see the        
c                             chapter l prelude).                       
c                                                                       
c   precision/hardware  - single and double/h32                         
c                       - single/h36,h48,h60                            
c                                                                       
c   reqd. imsl routines - ludatn,luelmn
c                                                                       
c   notation            - information on special notation and           
c                           conventions is available in the manual      
c                           introduction or through imsl routine uhelp  
c                                                                       
c   copyright           - 1978 by imsl, inc. all rights reserved.       
c                                                                       
c   warranty            - imsl warrants only that imsl testing has been 
c                           applied to this code. no other warranty,    
c                           expressed or implied, is applicable.        
c                                                                       
c-----------------------------------------------------------------------
c                                                                       
c                                                                       
      implicit real*8 (a-h,o-z)
c     implicit real (a-h,o-z)
      real*8             a(ia,n),b(n),wkarea(1000)  
c     real*4             a(ia,n),b(n),wkarea(400)                       
      data               zero/0.d0/,one/1.d0/                           
c                                  first executable statement           
c                                  lu decomposition of a                
      call ludatn (a,ia,n,a,ia,0,c1,c2,wkarea,wkarea,wa,ier)            
      if (d1 .lt. zero .and. ijob .ge. 1 .and. ijob .lt. 4) go to 5     
      d1 = c1                                                           
      d2 = c2                                                           
    5 if (ier .ge. 128) go to 60                                        
      if (ijob .le. 0 .or. ijob .gt. 4) go to 55                        
c                                  solve ax = b                         
      if (ijob .eq. 2 .or . ijob .eq. 3) call luelmn (a,ia,n,b,wkarea,b)
      if (ijob .ne. 1 .and. ijob .ne. 3) go to 9005                     
c                                  matrix inversion                     
      a(n,n) = one/a(n,n)                                               
      nm1 = n-1                                                         
      if (nm1 .lt. 1) go to 9005                                        
      do 40 ii=1,nm1                                                    
         l = n-ii                                                       
         m = l+1                                                        
         do 15 i=m,n                                                    
            sum = zero                                                  
            do 10 k=m,n                                                 
               sum = sum-a(i,k)*a(k,l)                                  
   10       continue                                                    
            wkarea(n+i) = sum                                           
   15    continue                                                       
         do 20 i=m,n                                                    
            a(i,l) = wkarea(n+i)                                        
   20    continue                                                       
         do 30 j=l,n                                                    
            sum = zero                                                  
            if (j .eq. l) sum = one                                     
            do 25 k=m,n                                                 
               sum = sum-a(l,k)*a(k,j)                                  
   25       continue                                                    
            wkarea(n+j) = sum/a(l,l)                                    
   30    continue                                                       
         do 35 j=l,n                                                    
            a(l,j) = wkarea(n+j)                                        
   35    continue                                                       
   40 continue                                                          
c                                  permute columns of a inverse         
      do 50 i=1,n                                                       
         j = n-i+1                                                     
c         jp = wkarea(j)      
         jp = int( wkarea(j) )                                                                                         
         if (j .eq. jp) go to 50                                        
         do 45 k=1,n                                                    
            c = a(k,jp)                                                 
            a(k,jp) = a(k,j)                                            
            a(k,j) = c                                                  
   45    continue                                                       
   50 continue                                                          
      go to 9005                                                        
   55 continue                                                          
c                                  warning with fix - ijob was set      
c                                  incorrectly                          
      ier = 65                                                          
      go to 9000                                                        
c                                  terminal error - matrix a is         
c                                  algorithmically singular             
   60 ier = 130                                                         
 9000 continue                                                          
c     call uertst(ier,'linv3f')                                         
 9005 return                                                            
      end                                                               
c
c
      subroutine ludatn (a,ia,n,xlu,ilu,idgt,d1,d2,apvt,equil,wa,ier)   
c
c   imsl routine name   - ludatn                                        
c                                                                       
c-----------------------------------------------------------------------
c                                                                       
c   computer            - prim77/single                                 
c                                                                       
c   latest revision     - june 1, 1982                                  
c                                                                       
c   purpose             - nucleus called only by imsl subroutine leqt2f 
c                                                                       
c   precision/hardware  - single and double/h32                         
c                       - single/h36,h48,h60                            
c                                                                       
c   reqd. imsl routines -       
c                                                                       
c   notation            - information on special notation and           
c                           conventions is available in the manual      
c                           introduction or through imsl routine uhelp  
c                                                                       
c   copyright           - 1982 by imsl, inc. all rights reserved.       
c                                                                       
c   warranty            - imsl warrants only that imsl testing has been 
c                           applied to this code. no other warranty,    
c                           expressed or implied, is applicable.        
c                                                                       
c-----------------------------------------------------------------------
c                                                                       
      implicit real*8 (a-h,o-z)
c     implicit real*4 (a-h,o-z)
c                                                                       
      real*8          a(ia,n),xlu(ilu,ilu),apvt(ilu),equil(n)           
c     real*4          a(ia,n),xlu(ilu,ilu),apvt(ilu),equil(n)           
      data  zero,one,four,sixtn,sixth/0.d0,1.d0,4.d0,16.d0,.0625d0/     
c                                  first executable statement           
c                                  initialization                       
      ier = 0                                                           
      rn = n                                                            
      wrel = zero                                                       
      d1 = one                                                          
      d2 = zero                                                         
      biga = zero                                                       
      do 10 i=1,n                                                       
         big = zero                                                     
         do 5 j=1,n                                                     
            p = a(i,j)                                                  
            xlu(i,j) = p                                                
c           p = dabs(p)                                                 
            p = abs(p)                                                 
            if (p .gt. big) big = p                                     
    5    continue                                                       
         if (big .gt. biga) biga = big                                  
         if (big .eq. zero) go to 110                                   
         equil(i) = one/big                                             
   10 continue                                                          
      do 105 j=1,n                                                      
         jm1 = j-1                                                      
         if (jm1 .lt. 1) go to 40                                       
c                                  compute u(i,j), i=1,...,j-1          
         do 35 i=1,jm1                                                  
            sum = xlu(i,j)                                              
            im1 = i-1                                                   
            if (idgt .eq. 0) go to 25                                   
c                                  with accuracy test                   
c           ai = dabs(sum)                                              
            ai = abs(sum)                                              
            wi = zero                                                   
            if (im1 .lt. 1) go to 20                                    
            do 15 k=1,im1                                               
               t = xlu(i,k)*xlu(k,j)                                    
               sum = sum-t                                              
c              wi = wi+dabs(t)                                          
               wi = wi+abs(t)                                          
   15       continue                                                    
            xlu(i,j) = sum                                              
c  20       wi = wi+dabs(sum)                                           
   20       wi = wi+abs(sum)                                           
            if (ai .eq. zero) ai = biga                                 
            test = wi/ai                                                
            if (test .gt. wrel) wrel = test                             
            go to 35                                                    
c                                  without accuracy                     
   25       if (im1 .lt. 1) go to 35                                    
            do 30 k=1,im1                                               
               sum = sum-xlu(i,k)*xlu(k,j)                              
   30       continue                                                    
            xlu(i,j) = sum                                              
   35    continue                                                       
   40    p = zero                                                       
c                                  compute u(j,j) and l(i,j), i=j+1,...,
         do 70 i=j,n                                                    
            sum = xlu(i,j)                                              
            if (idgt .eq. 0) go to 55                                   
c                                  with accuracy test                   
c           ai = dabs(sum)                                              
            ai = abs(sum)                                              
            wi = zero                                                   
            if (jm1 .lt. 1) go to 50                                    
            do 45 k=1,jm1                                               
               t = xlu(i,k)*xlu(k,j)                                    
               sum = sum-t                                              
c              wi = wi+dabs(t)                                          
               wi = wi+abs(t)                                          
   45       continue                                                    
            xlu(i,j) = sum                                              
c  50       wi = wi+dabs(sum)                                           
   50       wi = wi+abs(sum)                                           
            if (ai .eq. zero) ai = biga                                 
            test = wi/ai                                                
            if (test .gt. wrel) wrel = test                             
            go to 65                                                    
c                                  without accuracy test                
   55       if (jm1 .lt. 1) go to 65                                    
            do 60 k=1,jm1                                               
               sum = sum-xlu(i,k)*xlu(k,j)                              
   60       continue                                                    
            xlu(i,j) = sum                                              
c  65       q = equil(i)*dabs(sum)                                      
   65       q = equil(i)*abs(sum)                                      
            if (p .ge. q) go to 70                                      
            p = q                                                       
            imax = i                                                    
   70    continue                                                       
c                                  test for algorithmic singularity     
         q = rn+p                                                       
c   71    if (q .eq. rn) go to 110                                       
         if (q .eq. rn) go to 110                                       
         if (j .eq. imax) go to 80                                      
c                                  interchange rows j and imax          
         d1 = -d1                                                       
         do 75 k=1,n                                                    
            p = xlu(imax,k)                                             
            xlu(imax,k) = xlu(j,k)                                      
            xlu(j,k) = p                                                
   75    continue                                                       
         equil(imax) = equil(j)                                         
   80    apvt(j) = imax                                                 
         d1 = d1*xlu(j,j)                                               
c  85    if (dabs(d1) .le. one) go to 90                                
   85    if (abs(d1) .le. one) go to 90                                
         d1 = d1*sixth                                                  
         d2 = d2+four                                                   
         go to 85                                                       
c  90    if (dabs(d1) .ge. sixth) go to 95                              
   90    if (abs(d1) .ge. sixth) go to 95                              
         d1 = d1*sixtn                                                  
         d2 = d2-four                                                   
         go to 90                                                       
   95    continue                                                       
         jp1 = j+1                                                      
         if (jp1 .gt. n) go to 105                                      
c                                  divide by pivot element u(j,j)       
         p = xlu(j,j)                                                   
         do 100 i=jp1,n                                                 
            xlu(i,j) = xlu(i,j)/p                                       
  100    continue                                                       
  105 continue                                                          
c                                  perform accuracy test                
      if (idgt .eq. 0) go to 9005                                       
      p = 3*n+3                                                         
      wa = p*wrel                                                       
      q = wa+10.0**(-idgt)                                              
c  106 if (q .ne. wa) go to 9005                                         
      if (q .ne. wa) go to 9005                                         
      ier = 34                                                          
      go to 9000                                                        
c                                  algorithmic singularity              
  110 ier = 129                                                         
      d1 = zero                                                         
      d2 = zero                                                         
 9000 continue                                                          
c                                  print error                          
c     call uertst(ier,'ludatn')                                         
 9005 return                                                            
      end                                                               
c
c
      subroutine luelmn (a,ia,n,b,apvt,x)                               
c
c   imsl routine name   - luelmn                                        
c                                                                       
c-----------------------------------------------------------------------
c                                                                       
c   computer            - prim77/single                                 
c                                                                       
c   latest revision     - june 1, 1982                                  
c                                                                       
c   purpose             - nucleus called only by imsl subroutine leqt2f 
c                                                                       
c   reqd. imsl routines - none required                                 
c                                                                       
c   notation            - information on special notation and           
c                           conventions is available in the manual      
c                           introduction or through imsl routine uhelp  
c                                                                       
c   copyright           - 1982 by imsl, inc. all rights reserved.       
c                                                                       
c   warranty            - imsl warrants only that imsl testing has been 
c                           applied to this code. no other warranty,    
c                           expressed or implied, is applicable.        
c                                                                       
c-----------------------------------------------------------------------
c                                                                       
      implicit real*8 (a-h,o-z)
c     implicit real*4 (a-h,o-z)
c                                                                       
      real*8          a(ia,n),b(ia),apvt(ia),x(ia)                      
c     real*4          a(ia,n),b(ia),apvt(ia),x(ia)                      
c                                  first executable statement           
c                                  solve ly = b for y                   
      do 5 i=1,n                                                        
    5 x(i) = b(i)                                                       
      iw = 0                                                            
      do 20 i=1,n                                                       
c         ip = apvt(i) 
         ip = int( apvt(i) )                                                                                                    
         sum = x(ip)                                                    
         x(ip) = x(i)                                                   
         if (iw .eq. 0) go to 15                                        
         im1 = i-1                                                      
         do 10 j=iw,im1                                                 
            sum = sum-a(i,j)*x(j)                                       
   10    continue                                                       
         go to 20                                                       
   15    if (sum .ne. 0.) iw = i                                        
   20 x(i) = sum                                                        
c                                  solve ux = y for x                   
      do 30 ib=1,n                                                      
         i = n+1-ib                                                     
         ip1 = i+1                                                      
         sum = x(i)                                                     
         if (ip1 .gt. n) go to 30                                       
         do 25 j=ip1,n                                                  
            sum = sum-a(i,j)*x(j)                                       
   25   continue                                                        
   30 x(i) = sum/a(i,i)                                                 
      return                                                            
      end                                                               
c
c
c   imsl routine name   - mdnris
c 
c-----------------------------------------------------------------------
c 
c   computer            - cdcft5/single 
c 
c   latest revision     - november 1, 1984
c 
c   purpose             - inverse standard normal (gaussian)
c                           probability distribution function 
c 
c   usage               - call mdnris (p,y,ier) 
c 
c   arguments    p      - input value in the exclusive range (0.0,1.0)
c                y      - output value of the inverse normal (0,1)
c                           probability distribution function 
c                ier    - error parameter (output)
c                         terminal error
c                           ier = 129 indicates p lies outside the legal
c                             range. plus or minus machine infinity is
c                             given as the result (sign is the sign of
c                             the function value of the nearest legal 
c                             argument).
c 
c   precision/hardware  - single/all
c 
c   reqd. imsl routines - merfi 
c 
c   notation            - information on special notation and 
c                           conventions is available in the manual
c                           introduction or through imsl routine uhelp
c 
c   copyright           - 1982 by imsl, inc. all rights reserved. 
c 
c   warranty            - imsl warrants only that imsl testing has been 
c                           applied to this code. no other warranty,
c                           expressed or implied, is applicable.
c 
c-----------------------------------------------------------------------
c 
      subroutine mdnris (p,y,ier) 
c                                  specifications for arguments 
      real               p,y
      integer            ier
c                                  specifications for local variables 
      real               eps,g0,g1,g2,g3,h0,h1,h2,a,w,wi,sn,sd
      real               sigma,sqrt2,x,xinf 
c     data               xinf/.12650140831e+323/
      data               sqrt2/1.4142135623731/ 
      data               eps/.710543e-14/ 
      data               g0/.18511591e-3/,g1/-.20281520e-2/ 
      data               g2/-.14983844/,g3/.10786386e-1/
      data               h0/.99529751e-1/,h1/.52117329/ 
      data               h2/-.68883009e-1/
c                                  first executable statement 
c                                  first executable statement 
c Huge intrinsic isn't defined in F77 => just assign xinf to be a big number
c      xinf=huge(sigma)
      xinf=999999.99
      ier = 0 
      if (p .gt. 0.0 .and. p .lt. 1.0) go to 5
      ier = 129 
      sigma = sign(1.0,p) 
      y = sigma * xinf
      go to 9000
    5 if(p.le.eps) go to 10 
      x = 1.0 -(p + p)
      call merfi (x,y,ier)
      y = -sqrt2 * y
      go to 9005
c                                  p too small, compute y directly
   10 a = p+p 
      w = (-log(a+(a-a*a)))**.5
c                                  use a rational function in 1./w
      wi = 1./w 
      sn = ((g3*wi+g2)*wi+g1)*wi
      sd = ((wi+h2)*wi+h1)*wi+h0
      y = w + w*(g0+sn/sd)
      y = -y*sqrt2
      go to 9005
 9000 continue
 9005 return
      end 
c
c
      subroutine merfi (p,y,ier)
c
c   imsl routine name   - merfi 
c 
c---------------------------------------------------------------------- 
c 
c   computer            - cdcft5/single 
c 
c   latest revision     - january 1, 1978 
c 
c   purpose             - inverse error function
c 
c   usage               - call merfi (p,y,ier)
c 
c   arguments    p      - input value in the exclusive range (-1.0,1.0) 
c                y      - output value of the inverse error function
c                ier    - error parameter (output)
c                         terminal error
c                           ier = 129 indicates p lies outside the lega 
c                             range. plus or minus machine infinity is
c                             given as the result (sign is the sign of
c                             the function value of the nearest legal 
c                             argument).
c 
c   precision/hardware  - single/all
c 
c   notation            - information on special notation and 
c                           conventions is available in the manual
c                           introduction or through imsl routine uhelp
c 
c   copyright           - 1978 by imsl, inc. all rights reserved. 
c 
c   warranty            - imsl warrants only that imsl testing has been 
c                           applied to this code. no other warranty,
c                           expressed or implied, is applicable.
c 
c---------------------------------------------------------------------- 
c 
      implicit real(a-h,o-z)
c                                  specifications for arguments 
      real               p,y
      integer            ier
c                                  specifications for local variables 
      real               a(65),h1,h2,h3,h4,xinf,x,sigma,z,w,x3,x4,x5, 
     *                   x6,b 
      integer            n,ipp,l,lb2
      data               a(1),a(2),a(3),a(4),a(5),a(6),a(7),a(8),a(9),
     *                   a(10),a(11),a(12),a(13),a(14),a(15),a(16), 
     *                   a(17),a(18),a(19),a(20),a(21),a(22),a(23)
     *                   /.99288537662,.12046751614,
     *                   .16078199342e-01,.26867044372e-02, 
     *                   .49963473024e-03,.98898218599e-04, 
     *                   .20391812764e-04,.4327271618e-05,
     *                   .938081413e-06,.206734721e-06, 
     *                   .46159699e-07,.10416680e-07, 
     *                   .2371501e-08,.543928e-09,
     *                   .125549e-09,.29138e-10,
     *                   .6795e-11,.1591e-11, 
     *                   .374e-12,.88e-13,
     *                   .21e-13,.5e-14,
     *                   .1e-14/
      data               a(24),a(25),a(26),a(27),a(28),a(29),a(30), 
     *                   a(31),a(32),a(33),a(34),a(35),a(36),a(37), 
     *                   a(38),a(39),a(40)
     *                   /.91215880342e00,-.16266281868e-01,
     *                   .43355647295e-03,.21443857007e-03, 
     *                   .2625751076e-05,-.302109105e-05, 
     *                   -.12406062e-07,.62406609e-07,
     *                   -.540125e-09,-.142328e-08, 
     *                   .34384e-10,.33585e-10, 
     *                   -.1458e-11,-.81e-12, 
     *                   .53e-13,.2e-13,
     *                   -.2e-14/ 
      data               a(41),a(42),a(43),a(44),a(45),a(46),a(47), 
     *                   a(48),a(49),a(50),a(51),a(52),a(53),a(54), 
     *                   a(55),a(56),a(57),a(58),a(59),a(60),a(61), 
     *                   a(62),a(63),a(64),a(65)
     *                   /.95667970902,-.023107004309,
     *                   -.43742360975e-02,-.57650342265e-03, 
     *                   -.10961022307e-04,.25108547025e-04,
     *                   .10562336068e-04,.275441233e-05, 
     *                   .432484498e-06,-.20530337e-07, 
     *                   -.43891537e-07,-.1768401e-07,
     *                   -.3991289e-08,-.186932e-09,
     *                   .272923e-09,.132817e-09, 
     *                   .31834e-10,.1670e-11,
     *                   -.2036e-11,-.965e-12,
     *                   -.22e-12,-.1e-13,
     *                   .13e-13,.6e-14,
     *                   .1e-14/
      data               h1,h2,h3,h4/-1.5488130424, 
     *                   2.5654901231,-.55945763133,
     *                   2.2879157163/
c     data               xinf/.12650140831e+323/
c                                  first executable statement 
c Huge intrinsic isn't defined in F77 => just assign xinf to be a big number
c      xinf=huge(sigma)
      xinf=999999.99
      x = p 
      ier = 0 
      sigma = sign(1.,x)
      if (.not.(x.gt.-1..and.x.lt.1.)) go to 35 
      z = abs(x)
      if(z.gt..8) go to 20
      w = z*z/.32-1.
      n = 22
      ipp = 1 
      l = 1 
    5 lb2 = 1 
      x3 = 1. 
      x4 = w
      x6 = a(ipp) 
   10 x6 = x6 + a(ipp+lb2) * x4 
      x5 = x4 * w * 2.-x3 
      x3 = x4 
      x4 = x5 
      lb2 = lb2 + 1 
      if (lb2 .le. n) go to 10
      go to (15,30),l 
   15 y = z * x6 * sigma
      go to 9005
   20 b = (-log(1.-z*z))**.5 
      if (z .gt. .9975) go to 25
      w = h1*b+h2 
      ipp = 24
      l = 2 
      n = 16
      go to 5 
   25 w = h3 * b + h4 
      ipp = 41
      n = 24
      l = 2 
      go to 5 
   30 y = b * x6 * sigma
      go to 9005
   35 y = sigma*xinf
      ier = 129 
c 9000 continue
 9005 return
      end 
c
c

